#region Copyright And Revision History

/*---------------------------------------------------------------------------

	Caret.cs
	Copyright  2002 Bill Menees.  All rights reserved.
	Bill@Menees.com

	Who		When		What
	-------	----------	-----------------------------------------------------
	BMenees	10.27.2002	Created.

	BMenees	7.1.2003	Rewritten to more completely manage the caret.  Now
						it is much more like Petzold's example in "Programming
						Microsoft Windows with C#" (ch. 6, pg. 244).
						
	BMenees	6.4.2004	Fixed a bug where repeatedly setting Visible to the same
						value would keep calling ShowCaret or HideCaret, which 
						could cause it to stay visible or hidden too long.
-----------------------------------------------------------------------------*/

#endregion

using System;
using System.Drawing;
using System.Windows.Forms;
using System.Diagnostics;

namespace Menees
{
	/// <summary>
	/// Encapsulates the Win32 API for Carets.
	/// It inherits from MarshalByRefObject because it
	/// wraps unmanaged resources and can't be marshaled
	/// to remote contexts by value.
	/// </summary>
	public sealed class Caret : MarshalByRefObject, IDisposable
	{
		#region Public Members

		public Caret(Control Ctrl, int iHeight) : this(Ctrl, 2, iHeight)
		{
		}

		public Caret(Control Ctrl, int iWidth, int iHeight)
		{
			m_Ctrl = Ctrl;
			m_Size = new Size(iWidth, iHeight);
			m_Position = Point.Empty;

			Control.GotFocus += new EventHandler(ControlGotFocus);
			Control.LostFocus += new EventHandler(ControlLostFocus);

			//If the control already has focus, then create the caret.
			if (Ctrl.Focused)
			{
				ControlGotFocus(Ctrl, EventArgs.Empty);
			}
		}

		~Caret()
		{
			Dispose(false);
		}

		public void Dispose()
		{
			Dispose(true);
		}

		public Control Control
		{
			get
			{
				return m_Ctrl;
			}
		}

		public bool Visible
		{
			get
			{
				return m_bVisible;
			}
			set
			{
				if (m_bVisible != value)
				{
					m_bVisible = value;

					if (m_bCreatedCaret)
					{
						if (value)
						{
							Windows.ShowCaret(m_Ctrl);
						}
						else
						{
							Windows.HideCaret(m_Ctrl);
						}
					}
				}
			}
		}

		public Point Position
		{
			get
			{
				if (m_bCreatedCaret)
				{
					Windows.GetCaretPos(ref m_Position);
				}
				return m_Position;
			}
			set
			{
				if (m_bCreatedCaret)
				{
					Windows.SetCaretPos(value.X, value.Y);
				}
				else
				{
					m_Position = value;
				}
			}
		}

		public Size Size
		{
			get
			{
				return m_Size;
			}
		}

		#endregion

		#region Private Members

		private void Dispose(bool bDisposing)
		{
			if (bDisposing)
			{
				if (m_Ctrl.Focused)
				{
					ControlLostFocus(m_Ctrl, EventArgs.Empty);
				}

				m_Ctrl.GotFocus -= new EventHandler(ControlGotFocus);
				m_Ctrl.LostFocus -= new EventHandler(ControlLostFocus);

				GC.SuppressFinalize(this);
			}
			else if (m_bCreatedCaret)
			{
				//The GC called our Finalize method, so we only 
				//need to release unmanaged resources.
				Windows.DestroyCaret();
			}
		}

		private void ControlGotFocus(object sender, EventArgs e)
		{
			//Sometimes in the debugger, we'll get focus without ever having been sent
			//the LostFocus event.  So I'll fire it now just to make things balance out.
			//I'd rather make sure we clean up the old caret and then recreate a new one.
			//That seems safer than trying to reuse an existing one if it's in a state we
			//don't think should occur normally.
			ControlLostFocus(sender, e);

			m_bCreatedCaret = Windows.CreateCaret(m_Ctrl, m_Size.Width, m_Size.Height);
			if (m_bCreatedCaret)
			{
				Windows.SetCaretPos(m_Position.X, m_Position.Y);
				Visible = true;
			}
		}

		private void ControlLostFocus(object sender, EventArgs e)
		{
			if (m_bCreatedCaret)
			{
				Visible = false;
				Windows.DestroyCaret();
				m_bCreatedCaret = false;
			}
		}

		#endregion

		#region Private Data Members

		private Control m_Ctrl;
		private Size m_Size;
		private Point m_Position;
		private bool m_bVisible;
		private bool m_bCreatedCaret;

		#endregion		
	}
}
