﻿namespace Menees
{
    #region Using Directives

    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Text;

    #endregion

    /// <summary>
    /// Provides several methods for checking preconditions and postconditions
    /// (i.e., design-by-contract).
    /// </summary>
    public static class Conditions
    {
        #region Public Methods

        /// <summary>
        /// Makes sure the argument's state is valid (i.e., true).
        /// </summary>
        /// <param name="argState">The argument's state.</param>
        /// <param name="explanation">The name of the arg to put in the exception.</param>
        public static void RequireArgument(bool argState, string explanation)
        {
            RequireArgument(argState, explanation, null);
        }

        /// <summary>
        /// Makes sure the named argument's state is valid (i.e., true).
        /// </summary>
        /// <param name="argState">The argument's state.</param>
        /// <param name="explanation">The explanation to put in the exception.</param>
        /// <param name="argName">The name of the arg to put in the exception.</param>
        /// <exception cref="ArgumentException">If <paramref name="argState"/> is false.</exception>
        public static void RequireArgument(bool argState, string explanation, string argName)
        {
            if (!argState)
            {
                throw new ArgumentException(explanation, argName);
            }
        }

        /// <summary>
        /// Makes sure an object's Initialized flag is set to true.
        /// </summary>
        /// <param name="initialized">The initialized flag to check (e.g. this.initialized).</param>
        public static void RequireInitialized(bool initialized)
        {
            RequireState(initialized, "Object is not initialized.");
        }

        /// <summary>
        /// Makes sure a reference is non-null.
        /// </summary>
        /// <param name="reference">The reference to check.</param>
        /// <param name="argName">The arg name to put in the exception.</param>
        /// <exception cref="ArgumentNullException">If <paramref name="reference"/> is null.</exception>
        public static void RequireReference<T>(T reference, string argName)
            where T : class
        {
            if (reference == null)
            {
                throw new ArgumentNullException(argName);
            }
        }

        /// <summary>
        /// Makes sure the given state is valid (i.e., true).
        /// </summary>
        /// <param name="state">The state to check.</param>
        /// <param name="explanation">The explanation to put in the exception.</param>
        /// <exception cref="InvalidOperationException">If <paramref name="state"/> is false.</exception>
        public static void RequireState(bool state, string explanation)
        {
            if (!state)
            {
                throw new InvalidOperationException(explanation);
            }
        }

        /// <summary>
        /// Makes sure a string is non-null and non-empty.
        /// </summary>
        /// <param name="arg">The string to check.</param>
        /// <param name="argName">The name of the arg to put in the exception.</param>
        public static void RequireString(string arg, string argName)
        {
            RequireArgument(!string.IsNullOrEmpty(arg), "The string must be non-empty.", argName);
        }

        #endregion
    }
}
