#region Copyright And Revision History

/*---------------------------------------------------------------------------

	Copyright  2006 Bill Menees.  All rights reserved.
	Bill@Menees.com

	$History: VisualStudioInvoker.cs $
	
	*****************  Version 3  *****************
	User: Bill         Date: 1/16/12    Time: 9:09p
	Updated in $/CSharp/Menees/Classes
	Fixed a bug where the title bar icons were incorrect for MDI child
	windows.
	
	*****************  Version 2  *****************
	User: Bill         Date: 12/13/10   Time: 7:54a
	Updated in $/CSharp/Menees/Classes
	Updated to v3.0.0.
	
	*****************  Version 1  *****************
	User: Bill         Date: 7/08/06    Time: 1:58p
	Created in $/CSharp/Menees/Classes

-----------------------------------------------------------------------------*/

#endregion

#region Using Directives

using System;
using System.IO;
using System.Text;
using System.Diagnostics;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using System.Reflection;
using System.Threading;
using System.Globalization;

#endregion

namespace Menees
{
	/// <summary>
	/// This class is used to open a file in Visual Studio at a specific line number.
	/// Using Utilities.ShellExecute, it's very easy to open a file that's associated
	/// with Visual Studio.  But to force a file open in Visual Studio and then
	/// jump to a specific line number is a lot more work.
	/// <para/>
	/// The bulk of the code for this class was shamelessly pulled from FxCop 1.35's
	/// Microsoft.FxCop.UI.FxCopUI class (in FxCop.exe) using Reflector and the
	/// FileDisassembler add-in.
	/// </summary>
	public static class VisualStudioInvoker
	{
		#region Public Methods

		/// <summary>
		/// Opens a file in Visual Studio.
		/// </summary>
		/// <param name="fileName">The full path to a file to open.</param>
		/// <param name="fileLineNumber">The 1-based line number to go to in the file.</param>
		/// <returns>True if it was successful.  False if the file couldn't be opened in Visual Studio.</returns>
		public static bool OpenFile(string fileName, string fileLineNumber)
		{
			try
			{
				//Use late-bound COM to open the file in Visual Studio
				//so we can jump to a specific line number.  This also
				//allows us to reuse an open instance of VS.
				//
				//We could execute Visual Studio by command-line and
				//run the GotoLn command (like MegaBuild does), but that
				//requires starting a new instance of VS for each file opened.
				dynamic dte = GetVisualStudioInstance();
				OpenInVisualStudio(dte, fileName, fileLineNumber);
				return dte != null;
			}
			catch (COMException)
			{
				return false;
			}
			catch (ArgumentException)
			{
				return false;
			}
		}

		#endregion

		#region Private Methods

		private static object GetVisualStudioInstance()
		{
			object result = null;

			const string VisualStudioProgId = "VisualStudio.DTE";
			try
			{
				// See if there's a running instance of VS.
				result = Marshal.GetActiveObject(VisualStudioProgId);
			}
			catch (COMException)
			{
				// See if VS is registered.
				Type dteType = Type.GetTypeFromProgID(VisualStudioProgId, false);
				if (dteType != null)
				{
					// VS is registered, so we need to start a new instance.
					result = Activator.CreateInstance(dteType);
				}
			}

			return result;
		}

		private static void OpenInVisualStudio(dynamic dte, string fileName, string line)
		{
			dte.ExecuteCommand("File.OpenFile", Utilities.EnsureQuotes(fileName));
			dte.ExecuteCommand("Edit.Goto", line);

			dynamic mainWindow = dte.MainWindow;
			IntPtr mainWindowHandle = (IntPtr)Convert.ToInt64(mainWindow.HWnd);

			NativeMethods.BringWindowForward(mainWindowHandle);
			Thread.Sleep(1000);

			mainWindow.Activate();
			mainWindow.Visible = true;
		}

		#endregion

		#region Private Types

		private static class NativeMethods
		{
			public static void BringWindowForward(IntPtr hWnd)
			{
				if (IsIconic(hWnd))
				{
					const int SW_RESTORE = 9;
					ShowWindowAsync(hWnd, SW_RESTORE);
				}

				SetForegroundWindow(hWnd);
			}

			[return: MarshalAs(UnmanagedType.Bool)]
			[DllImport("user32.dll")]
			public static extern bool IsIconic(IntPtr hWnd);

			[return: MarshalAs(UnmanagedType.Bool)]
			[DllImport("user32.dll")]
			public static extern bool SetForegroundWindow(IntPtr hWnd);

			[return: MarshalAs(UnmanagedType.Bool)]
			[DllImport("user32.dll")]
			public static extern bool ShowWindowAsync(IntPtr hWnd, int iCmdShow);
		}

		#endregion
	}
}
